-- syntax
module Abstract where

import Util

import Control.Monad.Writer

import Data.Map (Map)
import qualified Data.Map as Map
import Data.Set (Set)
import qualified Data.Set as Set

type Name = String

data Sized = Sized | NotSized 
             deriving (Eq,Show)

data Co = Ind 
        | CoInd
          deriving (Eq,Show)

-- strict positivity 
data Pos 
  = SPos  -- strictly positive
  | NSPos -- unknown polarity
  | Neg   -- negative, used internally for contravariance of sized codata  
  | NOcc  -- non-occurring, used to determine whether data type is recursive
         deriving (Eq)

instance Show Pos where
  show SPos  = "+"
  show NSPos = "o"
  show Neg   = "-"
  show NOcc  = "T"

isSPos :: Pos -> Bool
isSPos SPos = True
isSPos NOcc = True
isSPos _ = False

polNeg :: Pos -> Pos
polNeg SPos = Neg
polNeg Neg = SPos
polNeg NSPos = NSPos
polNeg NOcc  = NOcc

-- polarity composition
polComp :: Pos -> Pos -> Pos
polComp NOcc  x   = NOcc
polComp x NOcc    = NOcc
polComp SPos  x   = x
polComp NSPos x   = NSPos
polComp Neg Neg   = SPos
polComp Neg NSPos = NSPos
polComp Neg SPos  = Neg

-- polarity conjuction (infimum)
polAnd :: Pos -> Pos -> Pos
polAnd NOcc x   = x
polAnd x NOcc   = x
polAnd NSPos x  = NSPos
polAnd x NSPos  = NSPos
polAnd SPos Neg = NSPos
polAnd Neg SPos = NSPos
polAnd SPos SPos = SPos
polAnd Neg Neg   = Neg

-- decoration of Pi-types
-- 1. whether argument is erased
-- further possibilities:
-- 2. polarity
-- 3. hidden

data Dec = Dec { erased :: Bool }
           deriving Eq

defaultDec = notErased
notErased = Dec False
argErased = Dec True
resurrectDec d = d { erased = False }

instance Show Dec where
    show dec = if erased dec then "[-]" else "(-)"

type MVar = Int -- metavariables are numbered

data Expr = Set
          -- size type
          | Size 
          | Succ Expr
          | Infty
          -- identifiers 
          | Meta MVar    -- meta-variable
          | Var Name     -- variables are named
          | Con Co Name  -- constructors
          | Def Name     -- fun/cofun ?
          | Let Name     -- definition (non-recursive)
          -- dependently typed lambda calculus
          | Case Expr [Clause]
          | LLet Name Expr Expr Expr --local let x : A = t in u
          | App Expr [Expr]
          | Lam Name Expr
          | Pi Dec Name Expr Expr
          deriving (Eq)

-- boilerplate to extract free variables in the usual sense
freeVars :: Expr -> Set Name
freeVars (Succ e) = freeVars e
freeVars (Var name) = Set.singleton name
freeVars (Case e cls) = Set.unions (freeVars e : map freeVarsClause cls) 
freeVars (LLet x a t u) = freeVars a `Set.union` (Set.delete x (freeVars t)) `Set.union` freeVars u
freeVars (App e es) = Set.unions (freeVars e : map freeVars es)
freeVars (Lam x e)  = Set.delete x (freeVars e)
freeVars (Pi _ x a b) = freeVars a `Set.union` Set.delete x (freeVars b)
freeVars _ = Set.empty

freeVarsClause :: Clause -> Set Name
freeVarsClause (Clause ps e) = foldl (Set.\\) (freeVars e) (map (Set.fromList . patternVars) ps)

patternVars :: Pattern -> [Name]
patternVars (VarP name) = [name]
patternVars (SuccP p)   = patternVars p
patternVars (ConP _ _ ps) = concat $ map patternVars ps
patternVars (DotP _) = []

-- get all the definitions that are refered to in expression
-- this is used e.g. to check whether a (co)fun is recursive
usedDefs :: Expr -> [Name]
usedDefs (Def n) = [n]
usedDefs (App e es) = foldl (++) (usedDefs e) (map usedDefs es)
usedDefs (Lam x e)  = usedDefs e
usedDefs (Pi _ _ a b) = usedDefs a ++ usedDefs b
usedDefs (LLet _ e e1 e2) = usedDefs e ++ usedDefs e1 ++ usedDefs e2
usedDefs (Succ e) = usedDefs e
usedDefs (Case e cls) = foldl (++) (usedDefs e) (map (usedDefs . clExpr) cls) 
usedDefs _ = []


instance Show Expr where
    show = prettyExpr

prettyExpr :: Expr -> String
prettyExpr e = 
    case e of
      Set -> "Set"
      Size -> "Size" 
      Succ e -> "($ " ++ prettyExpr e ++ ")"
      Infty -> "#"
      Meta i -> "?" ++ show i
      Var n -> n
      Con _ n -> n
      Def n -> n
      Let n -> n
      Case e cs -> "case " ++ prettyExpr e ++ " { " ++ prettyCases cs ++ " } "
      LLet n t1 e1 e2 ->
          "(let " ++ n ++ " : " ++ prettyExpr t1 ++ " = " ++ prettyExpr e1 ++ " in " ++ prettyExpr e2 ++ ")"  
      App e1 el -> "(" ++ prettyExprs (e1:el) ++ ")"
      Lam x e1 -> "(\\" ++ x ++ " -> " ++ prettyExpr e1 ++ ")"
      Pi dec x t1 t2 -> parens $ (if erased dec then brackets binding
                                    else if x=="" then s1 else parens binding) 
                                  ++ " -> " ++ prettyExpr t2
         where s1 = prettyExpr t1
               binding = if x == "" then s1 else x ++ " : " ++ s1 
               
prettyCase (Clause [p] e) = prettyPattern p ++ " -> " ++ prettyExpr e
prettyCases = prettyList "; " prettyCase 

prettyPattern :: Pattern -> String
prettyPattern = prettyExpr . patternToExpr
{-
prettyPattern (ConP c _ ps) = parens $ foldl (\ acc p -> acc ++ " " ++ prettyPattern p) c ps
prettyPattern (SuccP p) = parens $ "$ " ++ prettyPattern p
prettyPattern (DotP e)  = "." ++ prettyExpr e
prettyPattern (VarP x) = x
-}

prettyExprs :: [Expr] -> String
prettyExprs = prettyList " " prettyExpr


-- parallel substitution (CAUTION! NOT CAPTURE AVOIDING!) 

parSubst :: (Name -> Expr) -> Expr -> Expr
parSubst phi (Var x) = phi x
parSubst phi (Succ e) = Succ (parSubst phi e)
parSubst phi (App e es) = App (parSubst phi e) (map (parSubst phi) es)
parSubst phi (LLet x a b c) = LLet x (parSubst phi a) (parSubst phi b) (parSubst phi c)
parSubst phi (Pi dec x a b) = Pi dec x (parSubst phi a) (parSubst phi b)
parSubst phi (Lam x e)  = Lam x (parSubst phi e)
parSubst phi e = e

sgSubst :: Name -> Expr -> Expr -> Expr
sgSubst x t u = parSubst (\ y -> if x == y then t else Var y) u

-- metavariable substitution (CAUTION! NOT CAPTURE AVOIDING!) 

type Subst = Map MVar Expr

subst :: Subst -> Expr -> Expr
subst phi e@(Meta i) = 
  case Map.lookup i phi of
    Just e' -> e'
    _ -> e
subst phi (Succ e) = Succ (subst phi e)
subst phi (App e es) = App (subst phi e) (map (subst phi) es)
subst phi (LLet x a b c) = LLet x (subst phi a) (subst phi b) (subst phi c)
subst phi (Pi dec x a b) = Pi dec x (subst phi a) (subst phi b)
subst phi (Lam x e)  = Lam x (subst phi e)
subst phi e = e

-- declarations --------------------------------------------------

data Declaration 
  = DataDecl Name Sized Co [Pos] Telescope Type [Constructor]  -- data/codata
  | FunDecl Co [(TypeSig,[Clause])]  -- mutual fun block / mutual cofun block
  | LetDecl Bool TypeSig Expr -- bool for eval ??
    deriving (Eq,Show)

data TypeSig = TypeSig { namePart :: Name, typePart :: Type }
             deriving (Eq,Show)

type Type = Expr

type Constructor = TypeSig

type TBind = (Dec,Name,Type)  -- same as TypeSig ??

type Telescope = [TBind]

data Clause = Clause { clPatters :: [Pattern], clExpr :: Expr }  
            deriving (Eq,Show)

data Pattern = VarP Name
             | ConP Co Name [Pattern]
             | SuccP Pattern
             | UnusableP Pattern  
{- a pattern which results from matching a coinductive type and
the corresponding size index is not in the coinductive result type of
the function.  Such a pattern is not usable for termination
checking. -}
             | DotP Expr
               deriving (Eq,Show)

type Case = (Pattern,Expr)

-- Constructor analysis ----------------------------------------------

data FieldClass = Index 
                | Field (Maybe (Type, Clause)) deriving (Eq, Show)  
  -- maybe type and def of destructor
data FieldInfo = FieldInfo 
  { fDec :: Dec
  , fName  :: Name        -- "" for anonymous fields
  , fType  :: Type
  , fClass :: FieldClass 
  } deriving Show
data ConstructorInfo = ConstructorInfo 
  { cName   :: Name 
--  , cType   :: TVal
  , cFields :: [FieldInfo]
  , cTyCore :: Type
  , cPatFam :: (Bool, [Pattern]) 
  , cEtaExp :: Bool -- prerequisite: all destructors are defined
  } deriving Show

corePat :: ConstructorInfo -> [Pattern]
corePat = snd . cPatFam

{-  For the constructor

  vcons : (n : Nat) -> (head : A) -> (tail : Vec A n) -> Vec A (suc n)

cName   = "vcons"
-- cType   = evaluation of (A : Set) -> (n : Nat) -> ...
cFields = [("n",Nat,Index),("head",A,Field),("tail",Vec A n,Field)]
cTyCore = Vec A (suc n)
cPatFam = (True, [A, suc n])
cEtaExp = True, but may be set to False later since the constructor is recursive 

We generate the destructors
  
  head : (A : Set) -> (n : Nat) -> (x : Vec A (suc n)) -> A
  head .A .n (vcons A n _head _tail) = _head
  
  tail : (A : Set) -> (n : Nat) -> (x : Vec A (suc n)) -> Vec A n
  tail .A .n (vcons A n _head _tail) = _tail

in the implementation we use "constructed_by_head" for "x"

discriminate index arguments from fields  
  - split constructor type into telescope and core
    [(n : Nat),(head : A),(tail : Vec A n)], Vec A (suc n)
  - find free variables of core: [A,n]
  - create a list of (name,type,classification) for each constructor arg,
    where classification in {index,field}

-}

analyzeConstructor :: Co -> Telescope -> Constructor -> ConstructorInfo
analyzeConstructor co pars (TypeSig constrName ty) = 
  let (tele, core) = typeToTele ty
      fvs = freeVars core
      fields = map (classifyField fvs) tele
      fieldNames = map fName fields
      etaExpandable = all (/= "") fieldNames 
      -- generate destructors
      -- choose a name for the record to destroy
      indices = filter (\ f -> fClass f == Index) fields
      indexTele = map (\ f -> (fDec f, fName f, fType f)) indices
      indexNames = map fName indices
      destrNames = map fName $ filter (\ f -> fClass f /= Index) fields
      recName = "constructed_by_" ++ constrName
      parNames = map (\ (_,x,_) -> x) pars
      parAndIndexNames = parNames ++ indexNames
      -- substitute variable "fst" by application "fst A B p"
      phi x = if x `elem` destrNames 
                then App (Def x) (map Var (parAndIndexNames ++ [recName]))
                else Var x
      prefix d =  "destructor_argument_" ++ d
      -- modifiedDestrNames = map prefix destrNames
      -- TODO: Index arguments are not always before fields
      pattern = ConP co constrName (map (DotP . Var) parNames 
         ++ map (\ fi -> (case fClass fi of 
                            Index -> DotP . Var
                            _ -> VarP . prefix) 
                         (fName fi)) 
              fields)
      destrType t = teleToTypeErase (pars ++ indexTele) 
                      $ (Pi notErased recName core . parSubst phi) t
      destrBody (dn) = Clause (map VarP parAndIndexNames ++ [pattern]) (Var dn)
      fields' = if etaExpandable then
                  map (\ f -> if fClass f /= Index then
                                 f { fClass = Field $ Just 
                                        ( destrType (fType f)
                                        , destrBody (prefix (fName f)) )}
                               else f) fields
                 else fields 
  in ConstructorInfo 
       { cName = constrName
       , cFields = fields'
       , cTyCore = core
       , cPatFam = fromAllWriter $ isPatIndFamC core 
       , cEtaExp = etaExpandable
       } 
  where classifyField fvs (dec, name, ty) = FieldInfo
          { fDec = dec
          , fName  = name
          , fType  = ty
          , fClass = if name `Set.member` fvs then Index else Field Nothing
          }

-- Pattern inductive families ----------------------------------------

-- isPatIndFam takes a list of type signatures (constructor decls.)
-- and checks whether we have a pattern inductive family
-- in this case, a list of constructors with the associated
-- type indices (translated into pattern list) is returned
-- type parameters are dropped
{-
isPatIndFam :: Int -> [Constructor] -> Maybe [(Name,[Pattern])]
isPatIndFam numPars= mapM (\ tysig -> 
                             fmap (\ ps -> (namePart tysig, drop numPars ps))
                                  (isPatIndFamC (typePart tysig)))
-}

-- isPatIndFamC checks whether an expression (the type of s constructor) 
-- is of the form 
--   Gamma -> D ps
-- and returns the list ps of patterns if it is the case
isPatIndFamC :: Expr -> Writer All [Pattern]
isPatIndFamC (Def n) = return []
isPatIndFamC (App e es) = do
  ps  <- isPatIndFamC e
  ps' <- mapM exprToDotPat' es
  return $ ps ++ ps'
isPatIndFamC (Pi _ _ _ e) = isPatIndFamC e
isPatIndFamC _ = tell (All False) >> return []

-- Pattern auxiliary functions ---------------------------------------

-- for non-dot patterns, patterns overlap if one matches against the other
-- infinity size is represented as (DotP Infty)
-- I reprogram it here, since it does not need a monad
overlap :: Pattern -> Pattern -> Bool
overlap (VarP _) p' = True
overlap p (VarP _)  = True
overlap (ConP _ c ps) (ConP _ c' ps') = c == c' && overlaps ps ps'
-- size patterns always overlap
overlap (SuccP p) _ = True
overlap _ (SuccP p) = True
-- dot patterns always overlap (safe approximation)
overlap (DotP _) _ = True
overlap _ (DotP _) = True
{-
overlap (SuccP p) (SuccP p') = overlap p p'
overlap (SuccP p) (DotP Infty) = overlap p (DotP Infty)
overlap (DotP Infty) (SuccP p') = overlap (DotP Infty) p'
overlap (DotP Infty) (DotP Infty) = True
-}

overlaps :: [Pattern] -> [Pattern] -> Bool
overlaps ps ps' = and $ zipWith overlap ps ps'


-- exprToPattern is used to check whether an ind.fam. is a pattern ind.fam.
exprToPattern :: Expr -> Maybe Pattern
exprToPattern (Var n)    = return $ VarP n
exprToPattern (Con co n) = return $ ConP co n []
-- exprToPattern (Infty)    = return $ DotP Infty -- leads to non-term in compareExpr
exprToPattern (Succ e)   = exprToPattern e >>= return . SuccP
exprToPattern (App e es) = do
  p <- exprToPattern e
  ps <- mapM exprToPattern es
  p `patApps` ps
exprToPattern _ = fail "exprToPattern"

patApps :: (Monad m) => Pattern -> [Pattern] -> m Pattern
patApps p [] = return p
patApps (ConP co n ps) ps' = return $ ConP co n (ps ++ ps')
patApps _ _ = fail "patApps"


-- turn an expression into a pattern
-- the Bool is True if the pattern is proper, i.e., does not contain
-- DotP except (DotP Infty)
exprToDotPat :: Expr -> (Bool, Pattern)
exprToDotPat = fromAllWriter . exprToDotPat'

exprToDotPat' :: Expr -> Writer All Pattern
exprToDotPat' (Var n)    = return $ VarP n
exprToDotPat' (Con co n) = return $ ConP co n []
exprToDotPat' (Infty)    = return $ DotP Infty
exprToDotPat' (Succ e)   = exprToDotPat' e >>= return . SuccP
exprToDotPat' e@(App f []) = exprToDotPat' f
exprToDotPat' e@(App f es) = do
  p <- exprToDotPat' f
  case p of 
     (ConP co c ps) -> do ps' <- mapM exprToDotPat' es
                          return $ ConP co c (ps ++ ps')
     _ -> tell (All False) >> return (DotP e)
exprToDotPat' e = tell (All False) >> return (DotP e)

patternToExpr :: Pattern -> Expr
patternToExpr (VarP n)       = Var n
patternToExpr (ConP co n ps) = Con co n `App` (map patternToExpr ps)
patternToExpr (SuccP p)      = Succ (patternToExpr p)
patternToExpr (UnusableP p)  = patternToExpr p
patternToExpr (DotP e)       = e

completeP :: Pattern -> Bool
completeP (DotP _) = True
completeP (VarP _) = True
completeP (UnusableP p) = completeP p
completeP _ = False

isSuccessorPattern :: Pattern -> Bool
isSuccessorPattern (SuccP _) = True
isSuccessorPattern (DotP (Succ _)) = True
isSuccessorPattern _ = False

isVarP :: Pattern -> Bool
isVarP (VarP _) = True
isVarP (DotP (Var _)) = True
isVarP _ = False

shallowSuccP :: Pattern -> Bool
shallowSuccP (SuccP p) = isVarP p
shallowSuccP (DotP (Succ (Var _))) = True
shallowSuccP _ = False

rhsDefs :: [Clause] -> [Name]
rhsDefs cls = foldl (\ ns (Clause ps e) -> usedDefs e ++ ns) [] cls

teleToType' :: (Dec -> Dec) -> Telescope -> Type -> Type
teleToType' mod [] t = t
teleToType' mod ((dec,n,t):tel) t2 = Pi (mod dec) n t (teleToType' mod tel t2)

teleToType :: Telescope -> Type -> Type
teleToType = teleToType' id

teleToTypeErase :: Telescope -> Type -> Type
teleToTypeErase = teleToType' (\ dec -> dec { erased = True })

splitTeleType :: Int -> (Telescope,Type) -> (Telescope,Type)
splitTeleType 0 (tel,t) = (tel,t)
splitTeleType k (tel,(Pi dec n t t2)) = splitTeleType (k - 1) (tel ++ [(dec,n,t)],t2) 

-- | typeToTele ((x : A) -> (y : B) -> C) = ([(x,A),(y,B)], C)
typeToTele :: Type -> (Telescope, Type)
typeToTele t = ttt t []
    where 
      ttt :: Type -> Telescope -> (Telescope,Type)
      ttt (Pi dec n t' t2) tel = ttt t2 ((dec,n,t') : tel)
      ttt x tel = (reverse tel, x)                          
