module Eval where

import Util

import Abstract
import Value
import TCM

import TraceError

import Data.Maybe -- fromMaybe
import Data.List -- find
import Data.Map (Map)
import qualified Data.Map as Map

import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader

import Debug.Trace


traceEta msg a = a -- trace msg a 
traceEtaM msg = return () -- traceM msg

 

-- evaluation with rewriting -------------------------------------

{- 

Rewriting rules have the form

  blocked --> pattern

this means that at the root, at most one rewriting step is possible.
Rewriting rules are considered computational, since they trigger new
(symbolic) computations.  At least they have to be applied in

- pattern matching
- equality checking

When a new rule b --> p is added, b should be in --> normal form.
Otherwise there could be inconsistencies, like adding both rules

  b --> true
  b --> false

If after adding b --> true b is rewritten to nf, then the second rule
would be true --> false, which can be captured by MiniAgda.

Also, after adding a new rule, it could be used to rewrite the old rules.

Implementation:

- add a set of local rewriting rules to the context (not to the state)
- keep values in --> weak head normal form
- untyped equality test between values

 -}

reval :: Val -> TypeCheck Val
reval u = -- trace ("reval " ++ show u) $ 
 case u of
  VSet    -> return u
  VSize   -> return u
  VInfty  -> return u
  VSucc{} -> return u
{-
  VSucc v -> do v' <- reval v
                return $ succSize v'
 -}
  VApp v vl          -> do
    v'  <- reval v
    vl' <- mapM reval vl
    w   <- app v' vl'
    reduce w  -- since we only have rewrite rules at base types
              -- we do not need to reduces prefixes of w
  VCon{} -> return u
  VDef{} -> return u         -- CAN'T rewrite defined fun/data
  VGen{} -> reduce u  -- CAN rewrite variable

  VCase v env cl     -> do v' <- reval v
                           env' <- reEnv env
                           evalCase v' env' cl

  VPi dec x av env b -> do av' <- reval av
                           env' <- reEnv env
                           return $ VPi dec x av' env' b

  VLam x env e       -> do env' <- reEnv env
                           return $ VLam x env' e

  VMeta{} -> return u
  VUp v tv           -> do v' <- reval v
                           tv' <- reval tv
                           up v' tv'
  VIrr -> return u

reEnv :: Env -> TypeCheck Env
reEnv = mapM $ \ (x,v) -> do v' <- reval v
                             return (x,v')
   

-- reduce the root of a value
reduce :: Val -> TypeCheck Val
reduce v = -- trace ("reduce " ++ show v) $
 do
  rewrules <- asks rewrites
  mr <- findM (\ rr -> equal v (lhs rr)) rewrules
  case mr of
     Nothing -> return v
     Just rr -> trace ("firing " ++ show rr) $ return (rhs rr)

-- equal v v'  tests values for untyped equality
-- precond: v v' are in --> whnf
equal :: Val -> Val -> TypeCheck Bool
equal u1 u2 = -- trace ("equal " ++ show u1 ++ " =?= " ++ show u2) $
  case (u1,u2) of
    (v1,v2) | v1 == v2 -> return True
    (VSucc v1, VSucc v2) -> equal v1 v2  -- NO REDUCING NECC. HERE (Size expr)
    (VApp v1 vl1, VApp v2 vl2) -> 
       liftM2 (&&) (equal' v1 v2) (equals' vl1 vl2)
    (VPi dec1 x1 av1 env1 b1,VPi dec2 x2 av2 env2 b2) ->
       liftM2 (&&) (equal av1 av2) $     -- NO RED. NECC. (Type)
         new x1 (Domain av1 dec1) $ \ vx -> do 
                  v1 <- whnf (update env1 x1 vx) b1
                  v2 <- whnf (update env2 x2 vx) b2 
                  equal v1 v2

    (VLam x1 env1 b1, VLam x2 env2 b2) -> -- PROBLEM: DOMAIN MISSING
         addName x1 $ \ vx -> do          -- CAN'T "up" fresh variable
               do v1 <- whnf (update env1 x1 vx) b1
                  v2 <- whnf (update env2 x2 vx) b2
                  equal' v1 v2
    _ -> return False
    
equals' :: [Val] -> [Val] -> TypeCheck Bool
equals' [] []             = return True
equals' (v1:vs1) (v2:vs2) = liftM2 (&&) (equal' v1 v2) (equals' vs1 vs2)
equals' vl1 vl2           = return False

equal' = equal

{- LEADS TO NON-TERMINATION
-- equal' v1 v2  tests values for untyped equality
-- v1 v2 are not necessarily in --> whnf
equal' v1 v2 = do
  v1' <- reduce v1
  v2' <- reduce v2
  equal v1' v2'
-}

-- evaluation ----------------------------------------------------

-- weak head normal form
-- monadic, since it reads the globally defined constants from the signature
-- lets are expanded away

whnf :: Env -> Expr -> TypeCheck Val
whnf env e = --trace ("whnf " ++ show
  case e of
    Meta i -> return $ VMeta i 0
    LLet x t1 e1 e2 -> do v1 <- whnf env e1
                          whnf (update env x v1) e2
    Lam x e1 -> return $ VLam x env e1

    Pi dec x a b -> do 
      av <- whnf env a
      return $ VPi dec x av env b

    App e1 el -> do v1 <- whnf env e1
                    vl <- mapM (whnf env) el
                    app v1 vl
    Case e cs -> do v  <- whnf env e
                    evalCase v env cs
                  -- trace ("case head evaluates to " ++ showVal v) $ return ()

    Set -> return  VSet
    Infty -> return VInfty
    Succ e1 -> do v <- whnf env e1
                  return $ succSize v
                
    Size -> return VSize
    Con co n -> return $ VCon co n
    
    Def n -> return $ VDef n
    Let n -> do sig <- gets signature 
                let (LetSig _ e) = lookupSig n sig
                whnf [] e
    Var y -> lookupEnv env y


-- evaluate in standard environment
whnf' :: Expr -> TypeCheck Val
whnf' e = do
  env <- getEnv
  whnf env e

evalCase :: Val -> Env -> [Clause] -> TypeCheck Val
evalCase v env cs = do
  m  <- matchClauses env cs [v]
  case m of
    Nothing -> return $ VCase v env cs
    Just v' -> return $ v'

piApp :: TVal -> Val -> TypeCheck TVal
piApp (VPi dec x av env b) w = whnf (update env x w) b
-- piApp _ = fail "piApp: IMPOSSIBLE"

piApps :: TVal -> [Val] -> TypeCheck TVal
piApps tv [] = return tv
piApps tv (v:vs) = do tv' <- piApp tv v
                      piApps tv' vs

app :: Val -> [Val] -> TypeCheck Val
app u [] = return $ u
app u c = do  
         case (u,c) of
            (VApp u2 c2,_) -> app u2 (c2 ++ c)
            (VLam x env e,(v:vl))  -> do v' <- whnf (update env x v) e
                                         app v' vl
            (VDef n,_) -> appDef n c
            (VUp v (VPi dec x av rho b), w:wl) -> do
              v1 <- app v [w]  -- eta-expand w ??
              bv <- whnf (update rho x w) b
              v2 <- up v1 bv
              app v2 wl
            _ -> return $ VApp u c

-- unroll a corecursive definition one time (until constructor appears)
force :: Val -> TypeCheck Val
force v@ (VDef n) = --trace ("force " ++ show v) $
    do sig <- gets signature
       case lookupSig n sig of
         (FunSig CoInd t cl True) -> do m <- matchClauses [] cl []
                                        case m of 
                                          Just v' -> force v'
                                          Nothing -> return v
         _ -> return v
force v@(VApp (VDef n) vl) = --trace ("force " ++ show v) $
    do sig <- gets signature
       case lookupSig n sig of
         (FunSig CoInd t cl True) ->  do m <- matchClauses [] cl vl
                                         case m of 
                                           Just v' -> force v'
                                           Nothing -> return v
         _ -> return v
force v = return v

-- apply a recursive function
-- corecursive ones are not expanded even if the arity is exceeded 
-- this is because a coinductive type needs to be destructed by pattern matching
appDef :: Name -> [Val] -> TypeCheck Val
appDef n vl = --trace ("appDef " ++ n) $ 
    do
      sig <- gets signature
      case lookupSig n sig of
        (FunSig Ind _ cl True) -> do m <- matchClauses [] cl vl
                                     case m of
                                       Nothing -> return $ VApp (VDef n) vl
                                       Just v2 -> return v2
        _ -> return $ VApp (VDef n) vl   

-- reflection and reification  ---------------------------------------

up :: Val -> TVal -> TypeCheck Val
up v tv@(VPi dec x va rho b) = return $ VUp v tv
up v (VDef d) = upData v d []
up v (VApp (VDef d) vl) = upData v d vl
up v _ = return $ v

{- Most of the code to eta expand on data types is in 
   TypeChecker.hs "typeCheckDeclaration"

 Currently, eta expansion only happens at data *types* with exactly
one constructor.  In a first step, this will be extended to
non-recursive pattern inductive families.

The strategy is: match type value with result type for all the constructors
0. if there are no matches, eta expand to * (VIrr)
1. if there is exactly one match, eta expand accordingly using destructors
2. if there are more matches, do not eta-expand

up{Vec A (suc n)} x = vcons A n (head A n x) (tail A n x)

up{Vec Bool (suc zero)} x 
  = vcons Bool zero (head Bool zero x) (tail Bool zero x)
 
For vcons 
- the patterns of  Vec : (A : Set) -> Nat -> Set  are  [A,suc n]
- matching  Bool,suc zero  against  A,suc n  yields A=Bool,n=zero
- this means we can eta expand to vcons
- go through the fields of vcons
  - if Index use value obtained by matching
  - if Field destr, use  destr <all pars> <all indices> x
   
-}

-- eta expand  v  at data type  n vl
upData :: Val -> Name -> [Val] -> TypeCheck Val
upData v n vl = do -- trace ("upData " ++ show v ++ " at " ++ n ++ show vl) $ do
  let ret v' = traceEta ("Eta-expanding: " ++ show v ++ " --> " ++ show v' ++ " at type " ++ n ++ show vl) $ return v'
  sig <- gets signature
  case lookupSig n sig of
    (DataSig {symbTyp = dv, numPars = npars, isCo = co, constructors = cs, etaExpand = True}) -> if (null cs) then ret VIrr else do -- no constructor: empty type
      let (pars, inds) = splitAt npars vl
      -- for each constructor, match its core against the type
      -- produces a list of maybe (c.info, environment)
      cenvs <- mapM (\ ci -> let ps = snd . cPatFam $ ci
                             in liftM (fmap ((,) ci)) $ nonLinMatchList [] ps vl dv) cs
            
      -- traceM $ "Matching constructors: " ++ show cenvs
      case compressMaybes cenvs of
        -- exactly one matching constructor: can eta expand
        [(ci,env)] -> if not (cEtaExp ci) then return v else do
          -- get list of index values from environment
          let fis = cFields ci
          let indices = filter (\ fi -> fClass fi == Index) fis
          let indvs = map (\ fi -> fromJust $ lookup (fName fi) env) indices 
          let piv = pars ++ indvs ++ [v]
          -- for each argument of constructor, get value
          let arg (FieldInfo { fName = x, fClass = Index }) = 
                liftMaybe $ lookup x env 
              arg (FieldInfo { fName = d, fClass = Field _ }) = do
                -- lookup type sig  t  of destructor  d
                let FunSig { symbTyp = t } = lookupSig d sig
                -- pi-apply destructor type to parameters, indices and value v
                t' <- piApps t piv
                -- recursively eta expand  (d <pars> v)
                up (VDef d `VApp` piv) t'
          vs <- mapM arg fis 
          v' <- app (VCon co (cName ci)) (pars ++ vs) 
          ret v'
        -- more or less than one matching constructors: cannot eta expand
        l -> -- trace ("Eta: " ++ show (length l) ++ " patterns match at type " ++ show n ++ show vl) $ 
               return v
    _ -> return v


{-
      let matchC (c, ps, ds) = 
            do menv <- nonLinMatchList [] ps inds dv 
               case menv of
                 Nothing -> return False
                 Just env -> do
                   let grps = groupBy (\ (x,_) (y,_) -> x == y) env
                   -- TODO: now compare elements in the group
                   -- NEED types for equality check
                   -- trivial if groups are singletons
                   return $ all (\ l -> length l <= 1) grps
      cs' <- filterM matchC cs 
      case cs' of
        [] -> return $ VIrr
        [(c,_,ds)] ->  do
          let parsv = pars ++ [v]
          let aux d = do
               -- lookup type sig  t  of destructor  d
               let FunSig { symbTyp = t } = lookupSig d sig
               -- pi-apply destructor type to parameters and value v
               t' <- piApps t parsv
               -- recursively eta expand  (d <pars> v)
               up (VDef d `VApp` parsv) t'
          vs <- mapM aux ds 
          app (VCon co c) (pars ++ vs) 
        _ -> return v
    _ -> return v
-}
   
{- 
refl : [A : Set] -> [a : A] -> Id A a a
up{Id T t t'} x 
  Id T t t' =?= Id A a a  --> A = T, a = t, a = t'
-}
      
{- OLD CODE FOR NON-DEPENDENT RECORDS ONLY
    -- erase if n is a empty type
    (DataSig {constructors = []}) -> return $ VIrr
    -- eta expand v if n is a tuple type
    (DataSig {isCo = co, constructors = [c], destructors = Just ds}) -> do
       let vlv = vl ++ [v]
       let aux d = do -- lookup type sig  t  of destructor  d
                      let FunSig { symbTyp = t } = lookupSig d sig
                      -- pi-apply destructor type to parameters and value v
                      t' <- piApps t vlv
                      -- recursively eta expand  (d <pars> v)
                      up (VDef d `VApp` vlv) t'
       vs <- mapM aux ds 
       app (VCon co c) (vl ++ vs) -- (map (\d -> VDef d `VApp` (vl ++ [v])) ds)
    _ -> return v
END OLD CODE -}

-- pattern matching ---------------------------------------------------

matchClauses :: Env -> [Clause] -> [Val] -> TypeCheck (Maybe Val)
matchClauses env cl vl0 = do
  vl <- mapM reduce vl0  -- REWRITE before matching
  loop cl vl
    where loop [] vl = return Nothing
          loop (Clause pl rhs : cl2) vl = 
              do x <- matchClause env pl rhs vl 
                 case x of
                   Nothing -> loop cl2 vl
                   Just v -> return $ Just v

matchClause :: Env -> [Pattern] -> Expr -> [Val] -> TypeCheck (Maybe Val)
matchClause env [] rhs vl = do v <- whnf env rhs
                               v2 <- app v vl
                               return $ Just v2
matchClause env (p:pl) rhs (v:vl) =
  do m <- match env p v 
     case m of
       Just env' ->
         matchClause env' pl rhs vl
       Nothing -> return Nothing
-- too few arguments
matchClause _ _ _ [] = return Nothing

match :: Env -> Pattern -> Val -> TypeCheck (Maybe Env)
match env p v0 = --trace (show env ++ show v0) $ 
  do 
    -- force against coinductive constructor pattern
    v <- case p of 
           (ConP CoInd _ _) -> force v0
           _ -> return v0
    case (p,v) of
      (DotP _,_) -> return $ Just env
      (VarP x,_) -> return $ Just (update env x v)
      (ConP _ x [],VCon _ y) | x== y -> return $ Just env
      (ConP _ x pl,VApp (VCon _ y) vl) | x == y -> matchList env pl vl
      (SuccP p',VInfty) -> match env p' VInfty 
      (SuccP p',VSucc v') -> match env p' v'
      _ -> return Nothing

matchList :: Env -> [Pattern] -> [Val] -> TypeCheck (Maybe Env)
matchList env [] [] = return $ Just env
matchList env (p:pl) (v:vl) = do m <- match env p v 
                                 case m of 
                                   Just env' -> matchList env' pl vl
                                   Nothing -> return Nothing
matchList _ _ _ = return Nothing

-- typed non-linear matching -----------------------------------------

nonLinMatch :: Env -> Pattern -> Val -> TVal -> TypeCheck (Maybe Env)
nonLinMatch env p v0 tv = do 
    -- force against coinductive constructor pattern
      v <- case p of 
             (ConP CoInd _ _) -> force v0
             _ -> return v0
      case (p,v) of
        (DotP _,_) -> return $ Just env
        (VarP x,_) -> case find (\ (y,v') -> x == y) env of
                        Nothing -> return $ Just (update env x v)
                        Just (y,v') -> do 
                          b <- eqValBool tv v v' 
                          if b then return $ Just env else return Nothing
        (ConP _ c [],VCon _ c') | c == c' -> return $ Just env
        (ConP _ c pl,VApp (VCon _ c') vl) | c == c' -> do
           sy <- lookupSymb c
           nonLinMatchList env pl vl (symbTyp sy)
        (SuccP p',VInfty) -> nonLinMatch env p' VInfty tv
        (SuccP p',VSucc v') -> nonLinMatch env p' v' tv
        _ -> return Nothing


nonLinMatchList :: Env -> [Pattern] -> [Val] -> TVal -> TypeCheck (Maybe Env)
nonLinMatchList env [] [] tv = return $ Just env
nonLinMatchList env (p:pl) (v:vl) tv =
  case tv of
    VPi dec x va rho b -> do 
      m <- nonLinMatch env p v va 
      case m of 
         Just env' -> do
            vb <- whnf (update rho x v) b
            nonLinMatchList env' pl vl vb
         Nothing -> return Nothing
    _ -> fail $ "nonLinMatchList: cannot match in absence of pi-type"
nonLinMatchList _ _ _ _ = return Nothing


------------


eqValBool :: TVal -> Val -> Val -> TypeCheck Bool
eqValBool tv v v' = errorToBool $ eqVal tv v v'
-- eqValBool tv v v' = (eqVal tv v v' >> return True) `catchError` (\ _ -> return False)

eqVal :: TVal -> Val -> Val -> TypeCheck ()
eqVal = leqVal' N NSPos


-- force history
data Force = N | L | R -- not yet, left , right  
    deriving (Eq,Show)

switch :: Force -> Force
switch L = R
switch R = L
switch N = N

{-
eqVal' :: Force -> SemCxt -> Val -> Val -> TypeCheck ()
eqVal' f delta u1 u2 = --trace ("eqVal'" ++ show (u1,u2)) $ 
 do u1f <- force u1
    u2f <- force u2
    case (u1f /= u1,u2f /= u2) of
      (True,False) | f /= R -> 
                       eqVal' L delta u1f u2
      (False,True) | f /= L ->
                       eqVal' R delta u1 u2f 
      _ -> case (u1,u2) of
             (VSucc v1,VSucc v2) -> eqVal' f delta v1 v2
             (VApp v1 vl1,VApp v2 vl2) -> do eqVal' f delta v1 v2 
                                             eqVals' f delta vl1 vl2
             (VPi dec x1 av1 env1 b1,VPi dec x2 av2 env2 b2) -> 
                        do eqVal' f delta av1 av2
                           let (k, delta') = cxtPush dec av1 delta
                               new = VGen k
                           v1 <- whnf (update env1 x1 new) b1
                           v2 <- whnf (update env2 x2 new) b2 
                           eqVal' f delta' v1 v2

             (VLam x1 env1 b1, VLam x2 env2 b2) ->
                        do v1 <- whnf (update env1 x1 (VGen k)) b1
                           v2 <- whnf (update env2 x2 (VGen k)) b2
                           eqVal' f (k+1) v1 v2

             (v1,v2) | v1 == v2 -> return ()
             _ -> throwErrorMsg $ "eqVal error " ++ show f ++ ": " ++ show u1 ++ " != " ++ show u2

eqVals' :: Force -> SemCxt -> [Val] -> [Val] -> TypeCheck ()
eqVals' f delta [] [] = return ()
eqVals' f delta (v1:vs1) (v2:vs2) = do eqVal' f delta v1 v2 
                                       eqVals' f delta vs1 vs2
eqVals' f delta vl1 vl2 = throwErrorMsg "mismatch number of arguments"

eqVals :: SemCxt -> [Val] -> [Val] -> TypeCheck ()
eqVals = eqVals' N
-}


-- TODO: FOR THE FOLLOWING TO BE SOUND, ONE NEEDS COERCIVE SUBTYPING!
-- a function which does not use its argument can be used as just a function
-- [A] -> A <= A -> A
-- A <= [A]
leqErased :: Pos -> Dec -> Dec -> Bool
leqErased SPos  dec1 dec2 = erased dec2 || not (erased dec1)
leqErased Neg   dec1 dec2 = erased dec1 || not (erased dec2)
leqErased NSPos dec1 dec2 = erased dec1 == erased dec2 

-- subtyping

leqVal :: Pos -> TVal -> Val -> Val -> TypeCheck ()
leqVal = leqVal' N

leqVal' :: Force -> Pos -> TVal -> Val -> Val -> TypeCheck ()
leqVal' f p tv u1' u2' = {- do
  ce <- ask
  trace (("rewrites: " +?+ show (rewrites ce)) ++ "  leqVal': " ++ show ce ++ "\n |- " ++ show u1' ++ "\n <=" ++ show p ++ " " ++ show u2' ++ " : " ++ show tv ) $ 
  -}
   do u1 <- reduce u1'
      u2 <- reduce u2'
      u1f <- force u1
      u2f <- force u2
      case (u1f /= u1,u2f /= u2) of
        (True,False) | f /= R -> -- only unroll one side  
                     leqVal' L p tv u1f u2
        (False,True) | f /= L ->
                     leqVal' R p tv u1 u2f    
        _ -> case tv of
            VSize -> leqSize p u1 u2
            (VPi dec x av env b) -> 
{-
               do let (k, delta') = cxtPush dec av delta
                      new = VGen k
                  u1' <- app u1 [new]
                  u2' <- app u2 [new]
                  tv' <- whnf (update env x new) b
                  leqVal' f p delta' tv' u1' u2'           
-}
               new x (Domain av dec) $ \ gen -> do
                  u1' <- app u1 [gen]
                  u2' <- app u2 [gen]
                  tv' <- whnf (update env x gen) b
                  leqVal' f p tv' u1' u2'           

            _ -> case (u1,u2) of
              (VPi dec1 x1 av1 env1 b1, VPi dec2 x2 av2 env2 b2) ->
                 if not $ leqErased p dec2 dec1 then
                    throwErrorMsg $ "subtyping " ++ show u1 ++ " <=" ++ show p ++ " " ++ show u2 ++ " failed"
                  else do     
                    leqVal' (switch f) (polNeg p) VSet av1 av2
                    let (x, av) =  if p == Neg then (x1, av1) else (x2, av2)
                    new x (Domain av dec1) $ \ gen -> do
                      bv1 <- whnf (update env1 x1 gen) b1
                      bv2 <- whnf (update env2 x2 gen) b2
                      leqVal' f p VSet bv1 bv2
              (VApp v1 vl1,VApp v2 vl2) -> do
                           leqApp f p v1 vl1 v2 vl2
              (VCase v env cl, v2) -> leqCases (switch f) (polNeg p) tv v2 v env cl
              (v1, VCase v env cl) -> leqCases f p tv v1 v env cl
              (a1,a2) | a1 == a2 -> return ()
              _ -> leqApp f p u1 [] u2 []
                   
{- 
              (VSucc v1,VSucc v2) -> leqVal' f k v1 v2
              (VApp v1 vl1,VApp v2 vl2) -> do
                           leqApp f k v1 vl1 v2 vl2
              (VGen k1,VGen k2) -> if k1 == k2 then return () 
                                   else throwErrorMsg $ "gen mismatch "  ++ show k1 ++ " " ++ show k2 
              (VLam x1 env1 e1, VLam x2 env2 e2) ->
                  do v1 <- whnf (update env1 x1 (VGen k)) e1
                     v2 <- whnf (update env2 x2 (VGen k)) e2
                     leqVal' f (k+1) v1 v2
              _ ->  throwErrorMsg $ "leqVal error: " ++ show u1 ++ " != " ++ show u2
-}

leqCases :: Force -> Pos -> TVal -> Val -> Val -> Env -> [Clause] -> TypeCheck ()
leqCases f pol tv v1 v env cl = do
  vcase <- evalCase v env cl
  case vcase of 
    (VCase v env cl) -> mapM_ (leqCase f pol tv v1 v env) cl
    v2 -> leqVal' f pol tv v1 v2

leqCase :: Force -> Pos -> TVal -> Val -> Val -> Env -> Clause -> TypeCheck ()
leqCase f pol tv v1 v env (Clause [p] e) = 
  addPattern (dontCare `arrow` tv) p $ \ _ pv env' ->
    addRewrite (Rewrite v pv) [tv,v1] $ \ [tv',v1'] -> do
      v2  <- whnf (env' ++ env) e
      v2' <- reval v2
      leqVal' f pol tv' v1' v2'
   where dontCare = VSet -- (error "internal error in leqCase: unknown type")

leqVals' :: Force -> TVal -> [Pos] -> [Val] -> [Val] -> TypeCheck TVal
leqVals' f tv pos [] [] = return $ tv
leqVals' f (VPi dec x av env b) (p:pos) (v1:vs1) (v2:vs2) = 
  do when (not (erased dec)) $ leqVal' f p av v1 v2 
     tv' <- whnf (update env x (if p == Neg then v2 else v1)) b
     leqVals' f tv' pos vs1 vs2
leqVals' f tv pos vl1 vl2 = throwErrorMsg 
  "leqVals': not a function type or mismatch number of arguments"

{-
leqNe :: Force -> Val -> Val -> TypeCheck TVal
leqNe f v1 v2 = --trace ("leqNe " ++ show v1 ++ "<=" ++ show v2) $ 
  do case (v1,v2) of
      (VGen k1, VGen k2) -> if k1 == k2 then do 
                                 dom <- lookupGem k1
                                 return $ typ dom
                               else throwErrorMsg $ "gen mismatch "  ++ show k1 ++ " " ++ show k2 
-}
     
leqApp :: Force -> Pos -> Val -> [Val] -> Val -> [Val] -> TypeCheck ()
leqApp f pol v1 w1 v2 w2 = -- trace ("leqApp: " ++ show delta ++ " |- " ++ show v1 ++ show w1 ++ " <=" ++ show pol ++ " " ++ show v2 ++ show w2) $ 
  do case (v1,v2) of
      (VGen k1, VGen k2) -> 
        if k1 /= k2 
          then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2 
          else do tv <- lookupGenType k1
                  tv' <- leqVals' f tv (repeat pol) w1 w2
                  return ()
      (VCon _ n, VCon _ m) -> 
        if n /= m 
         then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2  
         else do 
          sig <- gets signature
          case lookupSig n sig of
            (ConSig tv) -> -- constructor
               leqVals' f tv (repeat NSPos) w1 w2 >> return ()

      (VDef n, VDef m) -> 
        if n /= m 
         then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2  
         else do 
          sig <- gets signature
          case lookupSig n sig of
            (DataSig{ numPars = p, positivity = pos, isSized = s, isCo = co, symbTyp = tv }) -> -- data type 
               let positivitySizeIndex = if s /= Sized then NSPos else 
                                           if co == Ind then SPos else Neg
                   pos' = -- trace ("leqApp:  posOrig = " ++ show (pos ++ [positivitySizeIndex])) $
                     map (polComp pol) (pos ++ positivitySizeIndex : repeat NSPos)
               in leqVals' f tv pos' w1 w2
                    >> return ()
{-
                  do
                    let params1 = (take p w1)
                    let (pparams1,nparams1) = posArgs params1 pos
                    let params2 = (take p w2)
                    let (pparams2,nparams2) = posArgs params2 pos
                    leqVals' f delta pparams1 pparams2 -- ok because parameters strictly pos
                    let indices1 = drop (p + 1) w1
                    let indices2 = drop (p + 1) w2
                    eqVals' f delta (nparams1 ++ indices1) (nparams2 ++ indices2) 
                    -- now one index (that may be a size) remains to be checked
                    case (length w2 > p && length w2 > p) of
                      True ->
                          case s of
                            Sized -> -- sized data type 
                                 case co of -- "substaging" on size index
                                   Ind ->
                                       leqSize (w1 !! p) (w2 !! p)
                                   CoInd ->
                                       leqSize (w2 !! p) (w1 !! p)
                            NotSized -> 
                                       eqVal delta (w1 !! p) (w2 !!p)
                      False -> return ()
-}

-- otherwise, we are dealing with a (co) recursive function
            (FunSig co tv _ _) ->  
               leqVals' f tv (repeat NSPos) w1 w2 >> return ()
                                                 
      _ -> throwErrorMsg $ "leqApp: " ++ show v1 ++ show w1 ++ " !<= " ++ show v2 ++ show w2
{-
             do leqVal delta v1 v2 
                eqVals delta w1 w2
 -}


-- substaging on size values
leqSize :: Pos -> Val -> Val -> TypeCheck ()
leqSize pol v1 v2 = -- trace ("leqSize " ++ show v1 ++ " <=" ++ show pol ++ " " ++ show v2) $ 
    do case (v1,v2) of
         (VSucc v1,VSucc v2) -> leqSize pol v1 v2
         (VGen v1,VGen v2) -> 
           if v1 == v2 then return ()
           else throwErrorMsg $ "leqSize: head mismatch: " ++ show v1 ++ " !<= " ++ show v2 
         (VInfty,VInfty) -> return ()
         _ -> case pol of
                SPos  -> leqSize' v1 v2
                NSPos -> leqSize' v1 v2 >> leqSize' v2 v1
                Neg   -> leqSize' v2 v1

leqSize' ::  Val -> Val -> TypeCheck ()  
leqSize' v1 v2 = --trace ("leqSize' " ++ show v1 ++ show v2) $ 
    do case (v1,v2) of
         (VSucc v1,VSucc v2) -> leqSize' v1 v2
         (VGen v1,VGen v2) -> 
           if v1 == v2 then return ()
           else throwErrorMsg $ "leqSize': head mismatch: " ++ show v1 ++ " !<= " ++ show v2 
         (_,VInfty) -> return ()
         (VMeta i n, VSucc v2) | n > 0 -> leqSize' (VMeta i (n-1)) v2
         (VMeta i n, VMeta j m) -> addLeq (VMeta i (n - min n m))
                                          (VMeta j (m - min n m)) 
         (VMeta i n, v2) -> addLeq v1 v2
         (VSucc v1, VMeta i n) | n > 0 -> leqSize' v1 (VMeta i (n-1))
         (v1,VMeta i n) -> addLeq v1 v2
         (v1,VSucc v2) -> leqSize' v1 v2
         _ -> throwErrorMsg $ "leqSize: " ++ show v1 ++ " !<= " ++ show v2
