{-# LANGUAGE TypeSynonymInstances #-}
 
module TCM where

import Util

import Abstract
import Value
import {-# SOURCE #-} Eval -- (up,whnf')

import TraceError

import Warshall

import Data.Map (Map)
import qualified Data.Map as Map

import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader

import Debug.Trace

-- reader monad for local environment (not used right now)

data TCContext = TCContext 
  { context  :: SemCxt
  , renaming :: Ren
  , environ  :: Env
  , rewrites :: Rewrites
  }

instance Show TCContext where
    show ce = show (environ ce) ++ "; " ++ show (context ce)

emptyContext = TCContext 
  { context  = cxtEmpty
  , renaming = Map.empty 
  , environ  = emptyEnv 
  , rewrites = emptyRewrites
  }

-- state monad for global signature

data TCState = TCState {
    signature   :: Signature,
    constraints :: Constraints
  }

type TypeCheck = StateT TCState (ReaderT TCContext (ErrorT TraceError IO)) 


-- rewriting rules -----------------------------------------------

data Rewrite  = Rewrite { lhs :: Val,  rhs :: Val }
type Rewrites = [Rewrite]

emptyRewrites = []

instance Show Rewrite where
  show rr = show (lhs rr) ++ " --> " ++ show (rhs rr)

-- renaming ------------------------------------------------------

type Ren = Map Name Int


{- context -------------------------------------------------------

A context maps generic values to their type value.

During type checking, named variables are mapped to
generic values via a renaming.  Thus, looking up the type of a
name involves first looking up the generic value, and then its type.

-}

data Domain = Domain { typ :: TVal, decor :: Dec } 

instance Show Domain where
  show item = (if erased (decor item) then brackets else id) (show (typ item))

data SemCxt = SemCxt { len :: Int, cxt :: Map Int TVal, decs :: Map Int Dec }
{- invariant: length (cxt delta) = length (decs delta) = len -}

instance Show SemCxt where
  show delta = show $ zipWith Domain (Map.elems (cxt delta))
                                     (Map.elems (decs delta))

cxtEmpty = SemCxt { len = 0, cxt = Map.empty, decs = Map.empty }

cxtPush' :: Dec -> TVal -> SemCxt -> SemCxt
cxtPush' dec v delta = delta { len = k + 1
                             , cxt  = Map.insert k v (cxt delta)
                             , decs = Map.insert k dec (decs delta) }
  where k = len delta

cxtPush :: Dec -> TVal -> SemCxt -> (Int, SemCxt)
cxtPush dec v delta = (len delta, cxtPush' dec v delta)

cxtPushGen ::  SemCxt -> (Int, SemCxt)
cxtPushGen delta = cxtPush bot bot delta
  where bot = error "IMPOSSIBLE: this generic is not bound to any type"

cxtSetType :: Int -> Dec -> TVal -> SemCxt -> SemCxt
cxtSetType k dec tv delta = delta { cxt  = Map.insert k tv (cxt delta)
                                  , decs = Map.insert k dec (decs delta) }

cxtLookupGen :: Monad m => SemCxt -> Int -> m Domain
cxtLookupGen delta k = do
  tv  <- liftMaybe $ Map.lookup k (cxt delta)
  dec <- liftMaybe $ Map.lookup k (decs delta)
  return $ Domain { typ = tv, decor = dec }

cxtLookupName :: Monad m => SemCxt -> Ren -> Name -> m Domain
cxtLookupName delta ren x = do
  i <- liftMaybe $ Map.lookup x ren
  cxtLookupGen delta i

-- clear all "erased" flags (see Pfenning, LICS 2001)
cxtResurrect :: SemCxt -> SemCxt
cxtResurrect delta = delta { decs = Map.map (\ _ -> Dec False) (decs delta) }
-- cxtResurrect delta = delta { decs = take (len delta) (repeat (Dec False)) }


-- manipulating the context ------------------------------------------

class Monad m => MonadCxt m where
--  bind     :: Name -> Domain -> Val -> m a -> m a
--  new performs eta-expansion "up" of new gen 
  newWithGen :: Name -> Domain -> (Int -> Val -> m a) -> m a   
  new        :: Name -> Domain -> (Val -> m a) -> m a
  new x d cont = newWithGen x d (\ _ -> cont)   
  new'       :: Name -> Domain -> m a -> m a
  new' x d cont = new x d (\ _ -> cont) 
  addName    :: Name -> (Val -> m a) -> m a   
--  addName x = new x dontCare
  setType    :: Int -> Domain -> m a -> m a
  lookupGen  :: Int -> m Domain
  lookupGenType :: Int -> m TVal
  lookupGenType i = lookupGen i >>= return . typ
  lookupName :: Name -> m Domain
  getLen     :: m Int       -- return length of the context
  getEnv     :: m Env       -- return current environment
  resurrect  :: m a -> m a -- resurrect all erased variables in context
  addRewrite :: Rewrite -> [Val] -> ([Val] -> m a) -> m a
  addPattern :: TVal -> Pattern -> (TVal -> Val -> Env -> m a) -> m a -- step under pat
  addPatterns:: TVal -> [Pattern] -> (TVal -> [Val] -> Env -> m a) -> m a 

dontCare = error "Internal error: tried to retrieve unassigned type of variable" 

instance MonadCxt TypeCheck where

  addName x f = do
    cxtenv <- ask
    let (k, delta) = cxtPushGen (context cxtenv)
    let v = VGen k
    let rho = update (environ cxtenv) x v 
    local (\ cxt -> cxt { context = delta
                        , renaming = Map.insert x k (renaming cxtenv)
                        , environ = rho }) (f v)

  newWithGen x dom f = do
    cxtenv <- ask
    let (k, delta) = cxtPush (decor dom) (typ dom) (context cxtenv)
    v <- up (VGen k) (typ dom)
    let rho = update (environ cxtenv) x v 
    local (\ cxt -> cxt { context = delta
                        , renaming = Map.insert x k (renaming cxtenv)
                        , environ = rho }) (f k v)

  setType k (Domain tv dec) = 
    local (\ ce -> ce { context = cxtSetType k dec tv (context ce) }) 

  lookupGen k = do
    ce <- ask
    cxtLookupGen (context ce) k

  lookupName x = do
    ce <- ask
    cxtLookupName (context ce) (renaming ce) x

  getLen = do
    ce <- ask
    return $ len (context ce)

  getEnv = do
    ce <- ask
    return $ environ ce

  resurrect cont = do
    ce <- ask
    local (\ _ -> ce { context = cxtResurrect (context ce) }) cont

  -- PROBABLY TOO INEFFICIENT
  addRewrite rew vs cont = trace ("adding rewrite " ++ show rew) $
    -- add rewriting rule
    local (\ cxt -> cxt { rewrites = rew : (rewrites cxt) }) $ do
      ce <- ask
      -- normalize all types in context
      trace "normalizing types in context" $ return ()
      cx' <- mapMapM reval (cxt (context ce))          -- LOOP!
      -- normalize environment 
      trace "normalizing environment" $ return ()
      en' <- reEnv (environ ce)
      -- normalize given values 
      vs' <- mapM reval vs
      -- continue in updated context
      local (\ ce -> ce { context = (context ce) { cxt = cx' }
                        , environ = en' }) $ cont vs' 

  -- addPattern :: TVal -> Pattern -> (TVal -> Val -> Env -> m a) -> m a
  addPattern (VPi dec x av env b) p cont =  
       case p of
          VarP y -> new y (Domain av dec) $ \ xv -> do
              bv <- whnf (update env x xv) b
              cont bv xv [(y,xv)]

          ConP co n pl -> do
              sig <- gets signature 
              cxt <- ask
              let (ConSig vc) = (lookupSig n sig)
              addPatterns vc pl $ \ vc' vpl rho -> do
                pv0 <- app (VCon co n) vpl
                pv  <- up pv0 av
                vb  <- whnf (update env x pv) b
                cont vb pv rho

          SuccP p2 -> do  
              cxt <- ask
              addPattern (VSize `arrow` VSize) p2 $ \ _ vp2 rho -> do
                let pv = succSize vp2 
                vb  <- whnf (update env x pv) b
                cont vb pv rho

{- Does NOT work, cannot be used for dot patterns
          DotP e -> do
              v  <- whnf' e
              vb <- whnf (update env x v) b
              cont vb xv []
-}
   
  addPatterns tv [] cont = cont tv [] []
  addPatterns tv (p:ps) cont = 
    addPattern tv p $ \ tv' v env -> 
      addPatterns tv' ps $ \ tv'' vs env' -> 
        cont tv'' (v:vs) (env' ++ env)

addBind :: Name -> Dec -> Expr -> TypeCheck a -> TypeCheck a
addBind x dec t cont = do
  tv <- whnf' t
  new' x (Domain tv dec) cont

-- signature -----------------------------------------------------
-- input to and output of the type-checker

type Signature = Map.Map Name SigDef

-- a signature entry is either  
-- * a fun/cofun, 
-- * a defined constant, 
-- * a constructor, or 
-- * a data type id with its kind
-- they share "symbTyp", the type signature of the definition
data SigDef 
  = FunSig  { isCo          :: Co 
            , symbTyp       :: TVal 
            , clauses       :: [Clause] 
            , isTypeChecked :: Bool 
            } --type , co , clauses , whether its type checked
  | LetSig  { symbTyp       :: TVal 
            , definingExpr  :: Expr 
            }-- type , expr 
  | ConSig  { symbTyp       :: TVal } -- type   
  | DataSig { numPars       :: Int
            , positivity    :: [Pos] 
            , isSized       :: Sized 
            , isCo          :: Co
            , symbTyp       :: TVal
            -- the following information is only needed for eta-expansion
            -- hence it is only provided for suitable ind.fams.
            , constructors  :: [ConstructorInfo]
            , etaExpand     :: Bool
{-
            , destructors   :: Maybe [Name] -- Nothing if not a record
            , isFamily      :: Bool 
-}
            } -- # parameters, positivity of parameters  , sized , co , type  
              deriving (Show)

emptySig = Map.empty

lookupSymb :: Name -> TypeCheck SigDef
lookupSymb n = do
  sig <- gets signature
  return $ lookupSig n sig 

lookupSig :: Name -> Signature -> SigDef
lookupSig n sig = case (Map.lookup n sig) of
                       Nothing -> error $ "Error not in signature: "  ++ n ++ " " ++ show sig
                       Just k -> k

addSig :: Name -> SigDef -> TypeCheck ()
addSig n def = do
  st <- get
  put $ st { signature = Map.insert n def $ signature st }

modifySig :: Name -> (SigDef -> SigDef) -> TypeCheck ()
modifySig n f = do
  st <- get
  put $ st { signature = Map.adjust f n $ signature st }

-- more on the type checking monad -------------------------------

initSt :: TCState
initSt = TCState emptySig emptyConstraints

resetConstraints :: TypeCheck ()
resetConstraints = do
  st <- get
  put $ st { constraints = emptyConstraints }

vGenSuccs (VGen k)  m = (k,m)
vGenSuccs (VSucc v) m = vGenSuccs v (m+1)

mkConstraint :: Val -> Val -> Maybe Constraint
mkConstraint (VMeta i n) (VMeta j m) = Just $ Arc (Flex i) (m-n) (Flex j)
mkConstraint (VMeta i n) VInfty      = Just $ Arc (Flex i) 0 (Rigid (RConst Infinite))
mkConstraint (VMeta i n) v           = Just $ Arc (Flex i) (m-n) (Rigid (RVar j))
  where (j,m) = vGenSuccs v 0
mkConstraint VInfty (VMeta i n)      = Just $ Arc (Rigid (RConst Infinite)) 0 (Flex i)
mkConstraint v (VMeta j m)           = Just $ Arc (Rigid (RVar i)) (m-n) (Flex j)
  where (i,n) = vGenSuccs v 0

-- addMeta k x  adds a metavariable which can refer to VGens < k
addMeta :: Int -> MVar -> TypeCheck ()
addMeta k i = do
  st <- get
  put $ st { constraints = NewFlex i (\ k' -> k' < k)
          -- DO NOT ADD constraints of form <= infty !!
          --               : Arc (Flex i) 0 (Rigid (RConst Infinite))
                         : constraints st }
 
addLeq :: Val -> Val -> TypeCheck ()
addLeq v1 v2 = -- trace ("Constraint: " ++ show v1 ++ " <= " ++ show v2) $
  case mkConstraint v1 v2 of
     Nothing -> return ()
     Just c -> do
       st <- get
       put $ st { constraints = c : constraints st }

solveConstraints :: TypeCheck Solution
solveConstraints = do
  cs <- gets constraints
  case solve cs of 
    Just subst -> -- trace ("solution" ++ show subst) $ 
                    return subst 
    Nothing    -> fail "size constraints unsolvable" 

-- solve constraints and substitute solution into the analyzed expression
solveAndModify :: Expr -> Env -> TypeCheck Expr
solveAndModify e rho = do
      sol <- solveConstraints
      let e' = subst (solToSubst sol rho) e
      resetConstraints
      return e'

nameOf :: Env -> Int -> Maybe Name
nameOf [] j = Nothing
nameOf ((x,VGen i):rho) j | i == j = Just x
nameOf (_:rho) j = nameOf rho j

solToSubst :: Solution -> Env -> Subst
solToSubst sol rho = Map.foldWithKey step Map.empty sol
  where step k (SizeVar i n) sub | Just x <- nameOf rho i =
           Map.insert k (add (Var x) n) sub
        step k (SizeConst Infinite) sub = Map.insert k Infty sub
        step _ _ sub = sub

        add e n | n <= 0 = e
                | otherwise = add (Succ e) (n-1)


-- pattern to Value ----------------------------------------------

patternToVal :: Pattern -> TypeCheck Val
patternToVal p = do
  k <- getLen
  return $ fst (p2v k p)

-- turn a pattern into a value
-- dot patterns get variables corresponding to their flexible generic value
p2v :: Int -> Pattern -> (Val,Int)
p2v k p = 
    case p of
      VarP n -> (VGen k,k+1)
      ConP co n [] -> (VCon co n,k)
      ConP co n pl -> let (vl,k') = ps2vs k pl
                      in (VApp (VCon co n) vl,k')
      SuccP p -> let (v,k') = p2v k p
                 in (VSucc v,k')   
      DotP e -> (VGen k,k+1)

ps2vs :: Int -> [Pattern] -> ([Val],Int)
ps2vs k []  = ([],k)
ps2vs k (p:pl) = let (v,k') = p2v k p
                     (vl,k'') = ps2vs k' pl
                 in
                   (v:vl,k'')
